import json
from collections import namedtuple

import pytest
import requests
from octokit import Octokit


class TestClientMethods(object):

    def test_client_methods_are_lower_case(self):
        for client in Octokit().__dict__:
            try:
                cls = getattr(Octokit(), client).__dict__
            except AttributeError:
                pass  # ignore non-class attributes
            assert all(method.islower() for method in cls)

    def test_method_has_doc_string(self):
        assert Octokit().authorization.get.__doc__ == 'Get a single authorization.'

    def test_method_has_name_string(self):
        assert Octokit().authorization.get.__name__ == 'get'

    def test_method_calls_requests(self, mocker):
        mocker.patch('requests.get')
        Octokit().authorization.get(id=1)
        assert requests.get.called
        assert requests.get.call_count == 1

    def test_has_required_method_parameters(self):
        with pytest.raises(AssertionError):
            Octokit().authorization.get()
        with pytest.raises(AssertionError):
            Octokit().authorization.get(id=None)

    def test_only_allows_valid_method_parameters(self):
        with pytest.raises(AssertionError):
            Octokit().authorization.get_grants(notvalid=1)

    def test_validate_method_parameters(self, mocker):
        mocker.patch('requests.get')
        Octokit().authorization.get(id=100)
        requests.get.assert_called_once_with(
            'https://api.github.com/authorizations/100', params={}, headers=Octokit().headers
        )

    def test_request_has_body_parameters(self, mocker):
        mocker.patch('requests.post')
        data = {'scopes': ['public_repo']}
        create = Octokit().authorization.create(**data)
        requests.post.assert_called_once_with(
            'https://api.github.com/authorizations', data=json.dumps(data), headers=create.headers
        )

    def test_must_include_required_body_parameters(self):
        data = {'gist_id': 'abc123'}
        with pytest.raises(AssertionError):
            Octokit().authorization.create(**data)

    def test_use_default_parameter_values(self, mocker):
        mocker.patch('requests.patch')
        headers = {'accept': 'application/vnd.github.squirrel-girl-preview', 'Content-Type': 'application/json'}
        data = {'state': 'open'}
        Octokit().issues.edit(owner='testUser', repo='testRepo', number=1)
        requests.patch.assert_called_once_with(
            'https://api.github.com/repos/testUser/testRepo/issues/1', data=json.dumps(data), headers=headers
        )

    def test_use_passed_value_instead_of_default_parameter_values(self, mocker):
        mocker.patch('requests.patch')
        headers = {'accept': 'application/vnd.github.squirrel-girl-preview', 'Content-Type': 'application/json'}
        data = {'state': 'closed'}
        Octokit().issues.edit(owner='testUser', repo='testRepo', number=1, **data)
        requests.patch.assert_called_once_with(
            'https://api.github.com/repos/testUser/testRepo/issues/1', data=json.dumps(data), headers=headers
        )

    def test_validate_enum_values(self):
        with pytest.raises(AssertionError):
            Octokit().issues.edit(owner='testUser', repo='testRepo', number=1, state='closeddddd')

    def test_non_default_params_not_in_the_url_for_get_requests_go_in_the_query_string(self, mocker):
        mocker.patch('requests.get')
        params = {'page': 2, 'per_page': '30'}
        Octokit().authorization.get_grant(id=404, page=2)
        requests.get.assert_called_once_with(
            'https://api.github.com/applications/grants/404', params=params, headers=Octokit().headers
        )

    def test_use_previous_values_if_available(self, mocker):
        mocker.patch('requests.patch')
        mocker.patch('requests.post')
        headers = {'accept': 'application/vnd.github.squirrel-girl-preview', 'Content-Type': 'application/json'}
        data = {'state': 'closed'}
        issue = Octokit().issues.edit(owner='testUser', repo='testRepo', number=1, **data)
        requests.patch.assert_called_with(
            'https://api.github.com/repos/testUser/testRepo/issues/1', data=json.dumps(data), headers=headers
        )
        issue2 = issue.issues.edit(**data)
        requests.patch.assert_called_with(
            'https://api.github.com/repos/testUser/testRepo/issues/1', data=json.dumps(data), headers=headers
        )
        issue2.pull_requests.create(head='branch', base='master', title='Title')
        requests.post.assert_called_with(
            'https://api.github.com/repos/testUser/testRepo/pulls',
            data=json.dumps({
                'base': 'master',
                'head': 'branch',
                'title': 'Title',
            }, sort_keys=True),
            headers={
                'Content-Type': 'application/json',
                'accept': 'application/vnd.github.machine-man-preview+json'
            }
        )

    def test_can_override_previous_values(self, mocker):
        mocker.patch('requests.patch')
        mocker.patch('requests.post')
        headers = {'accept': 'application/vnd.github.squirrel-girl-preview', 'Content-Type': 'application/json'}
        data = {'state': 'closed'}
        issue = Octokit().issues.edit(owner='testUser', repo='testRepo', number=1, **data)
        requests.patch.assert_called_with(
            'https://api.github.com/repos/testUser/testRepo/issues/1', data=json.dumps(data), headers=headers
        )
        issue.pull_requests.create(owner='user', head='branch', base='master', title='Title')
        requests.post.assert_called_with(
            'https://api.github.com/repos/user/testRepo/pulls',
            data=json.dumps({
                'base': 'master',
                'head': 'branch',
                'title': 'Title',
            }, sort_keys=True),
            headers={
                'Content-Type': 'application/json',
                'accept': 'application/vnd.github.machine-man-preview+json'
            }
        )

    def test_returned_object_is_self(self, mocker):
        mocker.patch('requests.patch')
        headers = {'accept': 'application/vnd.github.squirrel-girl-preview', 'Content-Type': 'application/json'}
        data = {'state': 'open'}
        sut = Octokit().issues.edit(owner='testUser', repo='testRepo', number=1)
        requests.patch.assert_called_once_with(
            'https://api.github.com/repos/testUser/testRepo/issues/1', data=json.dumps(data), headers=headers
        )
        assert sut.__class__.__name__ == 'Octokit'

    def test_returned_object_has_requests_response_object(self, mocker):
        patch = mocker.patch('requests.patch')
        Response = namedtuple('Response', ['json'])
        patch.return_value = Response(json=lambda: {})
        sut = Octokit().issues.edit(owner='testUser', repo='testRepo', number=1)
        assert sut._response.__class__.__name__ == 'Response'

    def test_returned_object_has_json_attribute(self, mocker):
        patch = mocker.patch('requests.patch')
        Request = namedtuple('Request', ['json'])
        patch.return_value = Request(json=lambda: data)
        data = {'documentation_url': 'https://developer.github.com/v3', 'message': 'Not Found'}
        sut = Octokit().issues.edit(owner='testUser', repo='testRepo', number=1)
        assert sut.json == data

    def test_returned_object_has_response_attributes(self, mocker):
        patch = mocker.patch('requests.patch')
        data = {
            "id": 1,
            "number": 1347,
            "state": "open",
            "title": "Found a bug",
            "user": {
                "login": "octocat",
                "site_admin": False
            },
            "labels": [{
                "id": 208045946,
            }, {
                "id": 208045947,
            }],
        }
        Request = namedtuple('Request', ['json'])
        patch.return_value = Request(json=lambda: data)
        sut = Octokit().issues.edit(owner='testUser', repo='testRepo', number=1)
        assert sut.response.id == 1
        assert sut.response.number == 1347
        assert sut.response.state == 'open'
        assert str(sut.response.user) == '<class \'octokit.ResponseData\'>'
        assert sut.response.user.login == 'octocat'
        assert sut.response.user.site_admin is False
        assert sut.response.labels[0].id == 208045946

    def test_returned_object_is_a_list(self, mocker):
        patch = mocker.patch('requests.patch')
        data = [{
            "id": 208045946,
        }, {
            "id": 208045947,
        }]
        Request = namedtuple('Request', ['json'])
        patch.return_value = Request(json=lambda: data)
        sut = Octokit().issues.edit(owner='testUser', repo='testRepo', number=1)
        assert sut.response[0].id == 208045946

    def test_an_exception_with_json_is_replaced_by_the_raw_text(self, mocker):
        patch = mocker.patch('requests.patch')
        Request = namedtuple('Request', ['json'])
        patch.return_value = Request(json=lambda: 'test')
        sut = Octokit().issues.edit(owner='testUser', repo='testRepo', number=1)
        assert sut.json == 'test'



import os
from collections import ChainMap
from collections import namedtuple

import pytest
import requests
from octokit import Octokit


class TestAuth(object):

    def test_can_set_basic_authentication(self):
        sut = Octokit(auth='basic', username='myuser', password='mypassword')
        assert sut.auth == 'basic'
        assert sut.username == 'myuser'
        assert sut.password == 'mypassword'

    def test_cannot_set_basic_authentication_with_out_required_data(self):
        with pytest.raises(KeyError):
            Octokit(auth='basic')
        with pytest.raises(AssertionError):
            Octokit(auth='basic', username='z', password='')
        with pytest.raises(KeyError):
            Octokit(auth='basic', password='mypassword')
        with pytest.raises(AssertionError):
            Octokit(auth='basic', username='')
        with pytest.raises(KeyError):
            Octokit(auth='basic', username='xyz')

    def test_basic_auth_used_if_set(self, mocker):
        mocker.patch('requests.get')
        Octokit(auth='basic', username='myuser', password='mypassword').authorization.get(id=100)
        requests.get.assert_called_once_with(
            'https://api.github.com/authorizations/100',
            params={},
            headers=Octokit().headers,
            auth=('myuser', 'mypassword')
        )

    def test_can_set_token_authentication(self):
        sut = Octokit(auth='token', token='yippy')
        assert sut.auth == 'token'
        assert sut.token == 'yippy'

    def test_cannot_set_token_authentication_with_out_required_data(self):
        with pytest.raises(KeyError):
            Octokit(auth='token')
        with pytest.raises(AssertionError):
            Octokit(auth='token', token='')

    def test_token_auth_used_if_set(self, mocker):
        mocker.patch('requests.get')
        Octokit(auth='token', token='yak').authorization.get(id=100)
        headers = dict(ChainMap(Octokit().headers, {'Authorization': 'token yak'}))
        requests.get.assert_called_once_with(
            'https://api.github.com/authorizations/100',
            params={},
            headers=headers,
        )

    def test_can_set_installation_authentication(self, mocker):
        Request = namedtuple('Request', ['json'])
        get = mocker.patch('requests.get')
        get.return_value = Request(json=lambda: [{'id': 13, 'app_id': 1}, {'id': 37, 'app_id': 42}])
        post = mocker.patch('requests.post')
        post.return_value = Request(json=lambda: {'token': 'v1.1f699f1069f60', 'expires_at': '2016-07-11T22:14:10Z'})
        with open(os.path.join(os.path.dirname(__file__), 'test.pem'), 'r') as f:
            private_key = f.read()
        sut = Octokit(auth='installation', app_id=42, private_key=private_key)
        assert sut.auth == 'installation'
        assert sut.token == 'v1.1f699f1069f60'
        assert sut.expires_at == '2016-07-11T22:14:10Z'

    def test_cannot_set_installation_authentication_with_out_required_data(self):
        with pytest.raises(KeyError):
            Octokit(auth='installation')
        with pytest.raises(AssertionError):
            Octokit(auth='installation', app_id='')
        with pytest.raises(KeyError):
            Octokit(auth='installation', app_id=42)
        with pytest.raises(AssertionError):
            Octokit(auth='installation', app_id=42, private_key='')

    def test_installation_token_is_used_if_set(self, mocker):
        Request = namedtuple('Request', ['json'])
        get = mocker.patch('requests.get')
        get.return_value = Request(json=lambda: [{'id': 13, 'app_id': 1}, {'id': 37, 'app_id': 42}])
        post = mocker.patch('requests.post')
        post.return_value = Request(json=lambda: {'token': 'v1.1f699f1069f60', 'expires_at': '2016-07-11T22:14:10Z'})
        with open(os.path.join(os.path.dirname(__file__), 'test.pem'), 'r') as f:
            private_key = f.read()
        sut = Octokit(auth='installation', app_id=42, private_key=private_key)
        assert sut.installation_id == 37
        get = mocker.patch('requests.get')
        sut.authorization.get(id=100)
        headers = {
            'Content-Type': 'application/json',
            'Authorization': 'token v1.1f699f1069f60',
            'accept': 'application/vnd.github.machine-man-preview+json'
        }
        requests.get.assert_called_once_with(
            'https://api.github.com/authorizations/100',
            params={},
            headers=headers,
        )

    def test_cannot_set_app_authentication_with_out_required_data(self):
        with pytest.raises(KeyError):
            Octokit(auth='app')
        with pytest.raises(AssertionError):
            Octokit(auth='app', app_id='')
        with pytest.raises(KeyError):
            Octokit(auth='app', app_id=42)
        with pytest.raises(AssertionError):
            Octokit(auth='app', app_id=42, private_key='')

    def test_can_set_app_authentication(self, mocker):
        Request = namedtuple('Request', ['json'])
        get = mocker.patch('requests.get')
        get.return_value = Request(json=lambda: [{'id': 37}])
        with open(os.path.join(os.path.dirname(__file__), 'test.pem'), 'r') as f:
            private_key = f.read()
        sut = Octokit(auth='app', app_id=42, private_key=private_key)
        assert sut.auth == 'app'
        assert sut.jwt is not None



import json

from octokit import webhook


class TestWebhook(object):

    def test_can_verify_webhook(self):
        headers = {
            'X-Hub-Signature': 'sha1=25af6174a0fcecc4d346680a72b7ce644b9a88e8',
            'X-GitHub-Event': 'push',
            'X-GitHub-Delivery': '72d3162f-cc78-11e3-81ab-4c9367dc0958'
        }
        payload = ''
        secret = 'secret'
        events = ['push']
        assert webhook.verify(headers, payload, secret, events=events)

    def test_can_filter_webhook_events(self):
        headers = {
            'X-Hub-Signature': 'sha1=5d61605c3feea9799210ddcb71307d4ba264225f',
            'X-GitHub-Delivery': '72d3162f-cc78-11e3-81ab-4c9367dc0958'
        }
        payload = ''
        secret = 'secret'
        events = ['push']
        assert webhook.verify(headers, payload, secret, events=events) is False

    def test_must_specify_events_to_allow(self):
        headers = {
            'X-Hub-Signature': 'sha1=5d61605c3feea9799210ddcb71307d4ba264225f',
            'X-GitHub-Delivery': '72d3162f-cc78-11e3-81ab-4c9367dc0958'
        }
        payload = ''
        secret = 'secret'
        assert webhook.verify(headers, payload, secret) is False

    def test_can_specify_all_events(self):
        headers = {
            'X-Hub-Signature': 'sha1=25af6174a0fcecc4d346680a72b7ce644b9a88e8',
            'X-GitHub-Event': 'push',
            'X-GitHub-Delivery': '72d3162f-cc78-11e3-81ab-4c9367dc0958'
        }
        payload = ''
        secret = 'secret'
        events = ['*']
        assert webhook.verify(headers, payload, secret, events=events)

    def test_only_known_events_are_valid(self):
        headers = {
            'X-Hub-Signature': 'sha1=5d61605c3feea9799210ddcb71307d4ba264225f',
            'X-GitHub-Event': 'pushy',
            'X-GitHub-Delivery': '72d3162f-cc78-11e3-81ab-4c9367dc0958'
        }
        payload = ''
        secret = 'secret'
        events = ['pushy']
        assert webhook.verify(headers, payload, secret, events=events) is False

    def test_delivery_guids_must_be_valid_guids(self):
        headers = {
            'X-Hub-Signature': 'sha1=5d61605c3feea9799210ddcb71307d4ba264225f',
            'X-GitHub-Event': 'push',
            'X-GitHub-Delivery': 'not-a-guid'
        }
        payload = ''
        secret = 'secret'
        events = ['push']
        assert webhook.verify(headers, payload, secret, events=events) is False

    def test_can_verify_user_agent(self):
        headers = {
            'X-Hub-Signature': 'sha1=25af6174a0fcecc4d346680a72b7ce644b9a88e8',
            'X-GitHub-Event': 'push',
            'X-GitHub-Delivery': '72d3162f-cc78-11e3-81ab-4c9367dc0958',
            'User-Agent': 'GitHub-Hookshot/',
        }
        payload = ''
        secret = 'secret'
        events = ['push']
        assert webhook.verify(headers, payload, secret, events=events, verify_user_agent=True)

    def test_verifies_user_agent(self):
        headers = {
            'X-Hub-Signature': 'sha1=25af6174a0fcecc4d346680a72b7ce644b9a88e8',
            'X-GitHub-Event': 'push',
            'X-GitHub-Delivery': '72d3162f-cc78-11e3-81ab-4c9367dc0958',
            'User-Agent': 'GitHub-Hooks',
        }
        payload = ''
        secret = 'secret'
        events = ['push']
        assert webhook.verify(headers, payload, secret, events=events, verify_user_agent=True) is False

    def test_verify_ping_event(self):
        headers = {
            'X-Hub-Signature': 'sha1=76b55589eeb1d5609a01922fc9a52475cf746a5b',
            'X-GitHub-Event': 'ping',
            'X-GitHub-Delivery': '72d3162f-cc78-11e3-81ab-4c9367dc0958',
            'User-Agent': 'GitHub-Hookshot/',
        }
        payload = json.dumps(
            {
                'hook': {
                    'type': 'App',
                    'id': 11,
                    'active': True,
                    'events': ['pull_request'],
                    'app_id': 42,
                }
            }
        )
        payload = '{"hook": {"events": ["pull_request"], "app_id": 42, "id": 11, "active": true, "type": "App"}}'
        secret = 'secret'
        app_id = webhook.verify(headers, payload, secret, events=['*'], return_app_id=True)
        assert app_id == 42

    def test_can_request_app_id_be_returned_on_non_ping_events(self):
        headers = {
            'X-Hub-Signature': 'sha1=25af6174a0fcecc4d346680a72b7ce644b9a88e8',
            'X-GitHub-Event': 'push',
            'X-GitHub-Delivery': '72d3162f-cc78-11e3-81ab-4c9367dc0958',
            'User-Agent': 'GitHub-Hookshot/',
        }
        payload = ''
        secret = 'secret'
        assert webhook.verify(headers, payload, secret, events=['*'], return_app_id=True)



import hashlib
import hmac
import json
from uuid import UUID

from octokit import utils


def valid_signature(headers, payload, secret):
    encoding = 'utf-8'
    algo, sig = headers.get('X-Hub-Signature').split('=')
    digest = hmac.new(secret.encode(encoding), payload.encode(encoding), getattr(hashlib, algo)).hexdigest()
    return hmac.compare_digest(sig.encode(encoding), digest.encode(encoding))


def valid_guid(guid):
    try:
        return str(UUID(guid)) == guid
    except ValueError:
        return False


def valid_event(event, events):
    return event in utils.get_json_data('events.json') and (event in events or '*' in events)


def valid_user_agent(ua):
    return ua.startswith('GitHub-Hookshot/')


def valid_headers(headers, events, verify_user_agent):
    if not valid_guid(headers.get('X-GitHub-Delivery')):
        return False
    if not valid_event(headers.get('X-GitHub-Event'), events):
        return False
    if verify_user_agent and not valid_user_agent(headers.get('User-Agent', '')):
        return False
    return True


def verify(headers, payload, secret, events=None, verify_user_agent=False, return_app_id=False):
    if not valid_headers(headers, events, verify_user_agent):
        return False
    validity = valid_signature(headers, payload, secret)
    if validity and return_app_id and headers.get('X-GitHub-Event') == 'ping':
        return json.loads(payload).get('hook').get('app_id')
    return validity



import datetime
import json
import re
from collections import ChainMap
from collections import defaultdict

import requests
from jose import jwt
from octokit import utils


class Base(object):

    headers = {'accept': 'application/vnd.github.v3+json', 'Content-Type': 'application/json'}
    base_url = 'https://api.github.com'

    def _get_headers(self, definition):
        return dict(ChainMap(definition.get('headers', {}), self.headers))

    def _validate(self, kwargs, params):
        cached_kwargs = dict(ChainMap(kwargs, self._attribute_cache['url']))
        required_params = [k for k, v in params.items() if v.get('required')]
        for p in required_params:
            assert p in cached_kwargs  # has all required
        for kwarg, value in kwargs.items():
            param_value = params.get(kwarg)
            assert param_value  # is a valid param but not necessarily required
            if param_value.get('enum'):
                assert value in param_value.get('enum')  # is a valid option of the enum
            if kwarg in required_params:
                assert value  # required param has a value

    def _form_url(self, values, _url, params):
        _values = dict(ChainMap(values, self._attribute_cache['url']))
        filtered_kwargs = {k: v for k, v in _values.items() if params.get(k)}
        data_values = filtered_kwargs.copy()
        for name, value in filtered_kwargs.items():
            _url, subs = re.subn(':{}'.format(name), str(value), _url)
            if subs != 0:
                self._attribute_cache['url'][name] = data_values.pop(name)
        url = '{}{}'.format(self.base_url, _url)
        return url, data_values

    def _get_data(self, kwargs, params):
        for param, value in params.items():
            if value.get('default') and not kwargs.get(param):
                kwargs[param] = value.get('default')
        return kwargs

    def _data(self, data_kwargs, params, method):
        data = self._get_data(data_kwargs, params)
        if method == 'get':
            return {'params': data}
        if method in ['post', 'patch', 'put', 'delete']:
            return {'data': json.dumps(data, sort_keys=True)}
        return {}

    def _setup_authentication(self, kwargs):
        authentication_schemes = {
            'basic': self._setup_basic_authentication,
            'token': self._setup_token_authentication,
            'installation': self._setup_installation_authentication,
            'app': self._setup_app_authentication,
        }
        if kwargs.get('auth'):
            authentication_schemes.get(kwargs.get('auth'))(kwargs)

    def _setup_basic_authentication(self, kwargs):
        assert kwargs['username']
        assert kwargs['password']
        self.username = kwargs['username']
        self.password = kwargs['password']
        self.auth = kwargs['auth']

    def _setup_token_authentication(self, kwargs):
        assert kwargs['token']
        self.token = kwargs['token']
        self.auth = kwargs['auth']

    def _setup_installation_authentication(self, kwargs):
        assert kwargs['app_id']
        assert kwargs['private_key']
        self.token, self.expires_at = self._app_auth_get_token(kwargs['app_id'], kwargs['private_key'])
        self.auth = kwargs['auth']
        self.headers['accept'] = 'application/vnd.github.machine-man-preview+json'

    def _setup_app_authentication(self, kwargs):
        assert kwargs['app_id']
        assert kwargs['private_key']
        self.jwt = self._app_auth_get_jwt(kwargs['app_id'], kwargs['private_key'])
        self.auth = kwargs['auth']
        self.headers['accept'] = 'application/vnd.github.machine-man-preview+json'

    def _app_auth_get_token(self, app_id, key):
        headers = {
            'Authorization': 'Bearer {}'.format(self._app_auth_get_jwt(app_id, key)),
            'Accept': 'application/vnd.github.machine-man-preview+json',
        }
        installation_url = '{}/app/installations'.format(self.base_url)
        installations = requests.get(installation_url, headers=headers).json()
        self.installation_id = [x.get('id') for x in installations if x.get('app_id') == app_id].pop()
        installation_token_url = '{}/installations/{}/access_tokens'.format(self.base_url, self.installation_id)
        response = requests.post(installation_token_url, headers=headers).json()
        return response['token'], response['expires_at']

    def _app_auth_get_jwt(self, app_id, key):
        payload = {
            'iat': int(datetime.datetime.timestamp(datetime.datetime.now())),
            'exp': int(datetime.datetime.timestamp(datetime.datetime.now())) + (10 * 60),
            'iss': app_id,
        }
        return jwt.encode(payload, key, algorithm='RS256')

    def _auth(self, requests_kwargs):
        if getattr(self, 'auth', None) == 'basic':
            return {'auth': (self.username, self.password)}
        if getattr(self, 'auth', None) in ['app', 'token', 'installation']:
            _headers = {
                'app': {
                    'Authorization': 'Bearer {}'.format(getattr(self, 'jwt', None))
                },
                'token': {
                    'Authorization': 'token {}'.format(self.token)
                },
                'installation': {
                    'Authorization': 'token {}'.format(self.token)
                },
            }
            headers = requests_kwargs['headers']
            headers.update(_headers.get(getattr(self, 'auth', None)))
            return {'headers': headers}
        return {}


class Octokit(Base):

    def __init__(self, *args, **kwargs):
        self._create(utils.get_json_data('rest.json'))
        self._setup_authentication(kwargs)
        self._attribute_cache = defaultdict(dict)

    def _create(self, definitions):
        for name, value in definitions.items():
            cls_name = utils.snake_case(name)
            client = self._create_client(cls_name, value)
            setattr(self, cls_name, client)

    def _create_client(self, name, methods):
        class_attributes = {}
        for _name, method in methods.items():
            method_name = utils.snake_case(str(_name))
            class_attributes.update({method_name: self._create_method(method_name, method)})
        bases = [object]
        cls = type(name, tuple(bases), class_attributes)
        return cls

    def _create_method(self, name, definition):

        def _api_call(*args, **kwargs):
            self._validate(kwargs, definition.get('params'))
            method = definition['method'].lower()
            requests_kwargs = {'headers': self._get_headers(definition)}
            url, data_kwargs = self._form_url(kwargs, definition['url'], definition.get('params'))
            requests_kwargs.update(self._data(data_kwargs, definition.get('params'), method))
            requests_kwargs.update(self._auth(requests_kwargs))
            _response = getattr(requests, method)(url, **requests_kwargs)
            try:
                attributes = _response.json()
            except ValueError:
                attributes = _response.text
            setattr(self, '_response', _response)
            setattr(self, 'json', attributes)
            setattr(self, 'response', self._convert_to_object(attributes))
            return self

        _api_call.__name__ = name
        _api_call.__doc__ = definition['description']
        return _api_call

    def _convert_to_object(self, item):
        if isinstance(item, dict):
            return type('ResponseData', (object, ), {k: self._convert_to_object(v) for k, v in item.items()})
        if isinstance(item, list):
            return list((self._convert_to_object(value) for index, value in enumerate(item)))
        else:
            return item
